\\ Copyright 2021, 2023 Kevin Ryde
\\
\\ This file is free software; you can redistribute it and/or modify it
\\ under the terms of the GNU General Public License as published by the Free
\\ Software Foundation; either version 3, or (at your option) any later
\\ version.
\\
\\ This file is distributed in the hope that it will be useful, but
\\ WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
\\ or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
\\ for more details.
\\
\\ You should have received a copy of the GNU General Public License along
\\ with this file  See the file COPYING.  If not, see
\\ <http://www.gnu.org/licenses/>.


default(strictargs,1);
default(recover,0);
read("../vecper.gp");


\\ Usage: gp <Hilbert-curve.gp
\\
\\ Hilbert's space filling curve
\\
\\     David Hilbert, "Ueber die stetige Abbildung einer Linie auf ein
\\     Flaechenstueck", Mathematische Annalen, volume 38, number 3, 1891,
\\     pages 459-460, DOI 10.1007/BF01199431
\\
\\ maps a fraction f in the range 0 <= f <= 1 to a location in the unit
\\ square,
\\
\\       1    +                    +
\\
\\    Y=3/4   5----6---7,9---10---11
\\            |         |          |       fractions n/16
\\    Y=1/2   4----3    8    13---12
\\                 |          |
\\    Y=1/4        2         14
\\                 |          |
\\     Y=0    0----1    +    15---16
\\
\\          X=0   1/4  1/2  3/4    1
\\
\\ The pattern of sub-blocks is
\\
\\            +---------+---------+
\\            |         |         |
\\            |    1    |    2    |
\\            | ------> | ------> |
\\            +---------+---------+
\\            | ^       |       | |
\\            | |  0    |    3  | |
\\            | |       |       v |
\\            +---------+---------+
\\
\\             ------------------>
\\
\\ Each part is a base 4 fractional digit from f.  Parts 1 and 2 are in the
\\ same orientation as the whole.  Part 0 is a transpose, and part 3 is a
\\ transpose and rotate 180 degrees.
\\
\\ In the code below, state "t" indicates a transpose required, and state
\\ "r" indicates rotate 180 degrees required.

real_to_frac(f) =
{
  if(type(f)=="t_REAL",
     \\ binary() gives [ [bits above point], [bits below point] ]
     \\ or 0.0 is just [0,...,0]
     my(v=binary(f));
     if(f,subst(Pol(v[2]),'x,2)) \\ fromdigits(), but work in gp 2.7.x
     / 2^#v[2];
     ,
     f);
}

\\ f is a rational 0 <= f <= 1.
\\ Return its Hilbert location as a complex rational x + I*y.
\\
digit_to_z_table = [0, I, 1+I, 2+I];
Hilbert_f_to_z_vecper(f) =
{
  f=real_to_frac(f);
  vecper_make_state
    ((state)->
     \\ print("state "state);
     my(f,t,r); [f,t,r]=state;
     (0 <= f && f <= 1) || error("oops, f out of range: ",f);
     f*=4;
     \\ print("  digit "d" from f="f);
     my(d=min(3,floor(f)));     \\ digit 0..3
     \\ print("  digit "d" from f="f);
     f-=d;
     my(z=digit_to_z_table[d+1]);
     if(t, z=I*conj(z));
     if(r, z=-z);
     if(d==0||d==3, t=!t);
     if(d==3, r=!r);
     \\ print("  return "z" ",[f,t,r]);
     [z, [f,t,r]];
     ,
     [f,0,0]);   \\ initial whole f, no transpose or rotate
}
Hilbert_f_to_z(f) = vecper_frac_fromdigits(Hilbert_f_to_z_vecper(f),2);

Hilbert_f_to_z(0)   == 0              || error();
Hilbert_f_to_z(1/4) == 1/2*I          || error();
Hilbert_f_to_z(1/2) == 1/2 + 1/2*I    || error();
Hilbert_f_to_z(3/4) == 1   + 1/2*I    || error();
Hilbert_f_to_z(1)   == 1              || error();
Hilbert_f_to_z(1-1/64) == 1 + 1/8*I   || error();

\\ The top left corner z = I of the unit square is by repeated 1 digits.
\\
vecper_frac_fromdigits([[],[1]],4) == 1/3      || error();
Hilbert_f_to_z(1/3) == I                       || error();

\\ The top right corner z = 1+I of the unit square is by repeated 2 digits.
\\
vecper_frac_fromdigits([[],[2]],4) == 2/3      || error();
Hilbert_f_to_z(2/3) == 1+I                     || error();

\\ The middle z = 1/2 + 1/2*I of the unit square has multiple visits, either
\\ f=1/2 between parts 1 and 2, or by corners of parts 0 and 3.
\\
Hilbert_f_to_z(1/2)         == 1/2 + 1/2*I     || error();
Hilbert_f_to_z((0 + 2/3)/4) == 1/2 + 1/2*I     || error();
Hilbert_f_to_z((3 + 1/3)/4) == 1/2 + 1/2*I     || error();

\\ There can be up to 4 visits to a binary corner point this way.
\\ For example in the above n/16 picture, point 1/2 + 1/4*I is a corner of 4
\\ surrounding quarters.
\\
Hilbert_f_to_z(( 1 + 2/3)/16) == 1/2 + 1/4*I   || error();
Hilbert_f_to_z(( 2 + 1/3)/16) == 1/2 + 1/4*I   || error();
Hilbert_f_to_z((13 + 2/3)/16) == 1/2 + 1/4*I   || error();
Hilbert_f_to_z((14 + 1/3)/16) == 1/2 + 1/4*I   || error();

\\ A point which is not a binary fraction (denominator a power of 2) is
\\ always between two successively subdivided squares, so is visited only
\\ twice (unless on the boundary in which case once).


\\-----------------------------------------------------------------------------

{
  print("plothraw binary points");
  my(len=4^6,
     \\ 0/len, 1/len, ..., len/len
     f_vec=vector(len+1,i, (i-1)/len),
     z_vec=apply(Hilbert_f_to_z, f_vec));
  plothraw(real(z_vec),imag(z_vec), 1);
}
{
  print("plothraw midpoints");
  my(len=4^6,
     \\ (0+1/2)/len, (1+1/2)/len, ..., (len-1+1/2)/len
     f_vec=vector(len,i, (i-1/2)/len),
     z_vec=apply(Hilbert_f_to_z, f_vec));
  z_vec=concat([0,1,1+I,I,0],z_vec);
  plothraw(real(z_vec),imag(z_vec), 1);
}

\\ {
\\   print("plothraw more");
\\   my(len=4^4,
\\      \\ (1/2)/len, (1+1/2)/len, ..., (len-1+1/2)/len
\\      f_vec=vector(len,i, (i-1 + 1/2)/len),
\\      z_vec=apply(Hilbert_f_to_z, f_vec));
\\   z_vec=concat([0,z_vec,[1+I]]);
\\   plothraw([1..#z_vec],imag(z_vec), 1);
\\ }

\\-----------------------------------------------------------------------------
