\\ Copyright 2019, 2020 Kevin Ryde
\\
\\ This file is free software; you can redistribute it and/or modify it
\\ under the terms of the GNU General Public License as published by the Free
\\ Software Foundation; either version 3, or (at your option) any later
\\ version.
\\
\\ This file is distributed in the hope that it will be useful, but
\\ WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
\\ or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
\\ for more details.
\\
\\ You should have received a copy of the GNU General Public License along
\\ with this file  See the file COPYING.  If not, see
\\ <http://www.gnu.org/licenses/>.


\\ Usage: gp <Peano-curve.gp
\\
\\     Guiseppe Peano, "Sur Une Courbe, Qui Remplit Toute Une Aire Plane",
\\     Mathematische Annalen, volume 36, number 1, 1890, pages 157-160.
\\     DOI 10.1007/BF01199438.
\\     http://www.springerlink.com/content/w232301n53960133/
\\
\\ Peano defines a mapping from a rational 0 <= f <= 1 to an X,Y location by
\\ taking f in ternary and extracting alternate digits to go to X and Y
\\ (like Z-order-curve.gp here), but applying a digit reversal 0<->2 when an
\\ odd number of 1-digits above it in the opposite parity positions.

default(recover,0);
default(strictargs,1);
read("../vecper.gp");

real_to_frac(f) =
{
  if(type(f)=="t_REAL",
     \\ binary() gives [ [bits above point], [bits below point] ]
     \\ or 0.0 is just [0,...,0]
     my(v=binary(f));
     if(f,subst(Pol(v[2]),'x,2)) \\ fromdigits(), but work in gp 2.7.x
     / 2^#v[2];
     ,
     f);
}

\\ f is a fraction 0 <= f <= 1.
\\ Peano_f_to_z_ternary() returns a vecper which is ternary digits for
\\ location z of f in the Peano curve.  Each "digit" is a complex integer
\\ x+I*y, 0 <= x,y <= 2.
\\
\\ vecper_make_state() is used to make ternary digits of x and y together in
\\ a complex number "digit".  Two ternary digits of f are taken to make one
\\ such complex digit, then continue on the rest of f likewise.
\\
\\ The "state" maintained for vecper_make_state() is
\\     [ f, xrev, yrev ]
\\ f is the fraction 0 <= f <= 1.  After taking its top-most two ternary
\\ digits, the remainder is scaled up by *9 so 0 <= f <= 1 again.
\\
\\ xrev is whether to apply the 0<->2 digit reversal to x.
\\ yrev is whether to apply the 0<->2 digit reversal to y.
\\ These go according to the number of 1 digits seen in their opposite.
\\ So xrev is 1-digits in y, and yrev is 1-digits in x.
\\ When an odd number of 1-digits, a reversal is wanted, hence types Mod(,2).
\\
\\ Beware that if denominator(f) doesn't have a factor of 3 then the length
\\ of the periodic part might be as long as denominator(f)-1.  The addhelp
\\ of vecper_frac_digits() describes this.  A floating point f is an n/2^k
\\ for k somewhere near default(realbitprecision), so potentially making a
\\ periodic part with length 2^96-1 which is hugely too big to create.
\\
Peano_f_to_z_ternary(f) =
{
  f=real_to_frac(f);
  \\ print("  "vecper_frac_digits(f,3));
  vecper_make_state
    ((state)-> \\ print("  "state);
               my(f=state[1],
                  xrev=state[2],
                  yrev=state[3]);
               (0 <= f && f <= 1) 
                 || error("Peano_f_to_z_ternary() oops, f out of range: ",f);
     
               f*=3;
               my(x=min(2,floor(f)));     \\ x digit 0 to 2
               f-=x;
               f*=3;
               my(y=min(2,floor(f)));     \\ y digit 0 to 2
               f-=y;
     
               if(xrev, x=2-x);           \\ possible reverse x
               xrev += y;
               yrev += x;
               if(yrev, y=2-y);           \\ possible reverse y
               [x + y*I, [f,xrev,yrev]];
               ,
               [f,Mod(0,2),Mod(0,2)]);   \\ initial whole f, no reversal
}

\\ f is a fraction 0 <= f <= 1.
\\ Return a complex number z which is its location in the Peano curve.
\\
Peano_f_to_z(f) = vecper_frac_fromdigits(Peano_f_to_z_ternary(f),3);


\\ f=0 is all digits 0 and no reversals so
\\
Peano_f_to_z_ternary(0) == [[],[0]]              || error();
Peano_f_to_z(0) == 0                             || error();
Peano_f_to_z(0.0) == 0                           || error();

\\ f=1 is all digits 2 and no reversals.

\\ In the Peano_f_to_z_ternary() code, f=1 would give floor(3*f)=3 as the
\\ "digit" for x, but limiting to digit 0 <= x,y <= 2, means effectively
\\ using representation f = 1 = 0.222..., and the ternary output similarly.
\\ This is like vecper_frac_digits() flag=1 "nines" option does for ordinary
\\ fractions.
\\
Peano_f_to_z_ternary(1) == [[],[2+2*I]]          || error();
Peano_f_to_z(1) == 1 + I                         || error();

\\ f=1/2 is the middle of the unit square 0 to 1+I, so x=1/2 and y=1/2
\\
Peano_f_to_z_ternary(1/2) == [[], [1+I,1+I]]     || error();
Peano_f_to_z(1/2) == 1/2 + 1/2*I                 || error();

\\ top left corner x=0,y=1 is f = .020202... = 1/4
\\
vecper_frac_fromdigits([[], [0,2]],3) == 1/4     || error();
Peano_f_to_z_ternary(1/4) == [[], [2*I]]         || error();
Peano_f_to_z(1/4) ==  0  + I                     || error();

\\ bottom right corner x=1,y=0 is conversely f = .202020... = 3/4
\\
vecper_frac_fromdigits([[], [2,0]],3) == 3/4     || error();
Peano_f_to_z_ternary(3/4) == [[], [2]]           || error();
Peano_f_to_z(3/4) ==  1                          || error();

\\ f=1/3 = .10000 has a digit x=1 which flips all the y digits below it so
\\
Peano_f_to_z_ternary(1/3) == [[1+2*I], [2*I]]    || error();
Peano_f_to_z(1/3) == 1/3 + I                     || error();


\\ The effect of the digit flips is to traverse the unit square x=0,y=0 to
\\ x=1,y=1 by sub-squares corner-to-corner.
\\
\\ The first ternary digit of f below the radix point goes to x, so x is in
\\ the range 0 to 1/3 while tne lower digits run y up from 0 to 1.  Next x
\\ is 1/3 to 2/3 and y runs down 1 to 0, and finally x is 2/3 to 1 and y up
\\ from 0 to 1.  The latter is a repeat of the first column.
\\
\\     +---------+----------+---------+
\\     |     3/9 | 3/9      |      9/9|
\\     |    /    |     \    |    /    |
\\     |   /     |      \   |   /     |
\\     |2/9 -----+----- 4/9 + 8/9 ----+
\\     |   \     |     /    |   \     |
\\     |    \    |    /     |     \   |
\\     +---- 1/9 + 5/9 -----+----- 7/9|
\\     |    /    |    \     |     /   |
\\     |   /     |     \    |   /     |
\\     |0/9      |      6/9 | 6/9     |
\\     +---------+----------+---------+
\\
\\ It can be noted 1/9 and 5/9 are the same location, and 4/9 and 8/9
\\ similarly back from the end.
\\
Peano_f_to_z(1/9) == 1/3 + 1/3*I                     || error();
Peano_f_to_z(5/9) == 1/3 + 1/3*I                     || error();

Peano_f_to_z(4/9) == 2/3 + 2/3*I                     || error();
Peano_f_to_z(8/9) == 2/3 + 2/3*I                     || error();

\\ The other inside corners where 4 squares meet are the 1/4 and 3/4 offsets
\\ for "top left" etc in their sub-squares.
\\
(1 + 1/4)/9 == 5/36                    || error();
(2 + 3/4)/9 == 11/36                   || error();
(3 + 1/4)/9 == 13/36                   || error();
(4 + 3/4)/9 == 19/36                   || error();
Peano_f_to_z( 5/36) == 1/3 + 2/3*I     || error();
Peano_f_to_z(11/36) == 1/3 + 2/3*I     || error();
Peano_f_to_z(13/36) == 1/3 + 2/3*I     || error();
Peano_f_to_z(19/36) == 1/3 + 2/3*I     || error();

(4 + 1/4)/9 == 17/36                   || error();
(5 + 3/4)/9 == 23/36                   || error();
(6 + 1/4)/9 == 25/36                   || error();
(7 + 3/4)/9 == 31/36                   || error();
Peano_f_to_z(17/36) == 2/3 + 1/3*I     || error();
Peano_f_to_z(23/36) == 2/3 + 1/3*I     || error();
Peano_f_to_z(25/36) == 2/3 + 1/3*I     || error();
Peano_f_to_z(31/36) == 2/3 + 1/3*I     || error();


\\-----------------------------------------------------------------------------
\\ Plot fractal using Peano_f_to_z() for coordinates.
\\ len is resolution steps 1/len to plot.
\\ If len is a power-of-3 then diagonal crossings of squares is seen.
\\
\\ If len not a power-of-3 then the fractional locations are a wilder
\\ selection of points within sub-squares.  The overall shape is still a
\\ square, but the inside steps are a bit strange.
\\
{
  print("plothraw some points");
  my(len=700,                            \\ or try 729 for diagonal grid
     f_vec=vector(len+1,i, (i-1)/len),   \\ 0, 1/len, ..., len/len (=1)
     z_vec=apply(Peano_f_to_z, f_vec));
  plothraw(real(z_vec),imag(z_vec), 1);
}

\\ Taking offsets 1/2 into each 1/3^k sub-square is the middle of each
\\ square, and gives the pattern oscillating back and forward.
\\ This is the most common form of the curve in integers, though the
\\ diagonals might be thought more fundamental and these its midpoint curve.
{
  print("plothraw midpoints");
  my(len=3^8,
     \\ (1/2)/len, (1+1/2)/len, ..., (len-1+1/2)/len
     f_vec=vector(len,i, (i-1 + 1/2)/len),
     z_vec=apply(Peano_f_to_z, f_vec));
  z_vec=concat([0,z_vec,[1+I]]);
  plothraw(real(z_vec),imag(z_vec), 1);
}

{
  print("plothraw more");
  my(len=3^8,
     \\ (1/2)/len, (1+1/2)/len, ..., (len-1+1/2)/len
     f_vec=vector(len,i, (i-1 + 1/2)/len),
     z_vec=apply(Peano_f_to_z, f_vec));
  z_vec=concat([0,z_vec,[1+I]]);
  plothraw([1..#z_vec],imag(z_vec), 1);
}

\\-----------------------------------------------------------------------------
