\\ Copyright 2019 Kevin Ryde
\\
\\ This file is free software; you can redistribute it and/or modify it
\\ under the terms of the GNU General Public License as published by the Free
\\ Software Foundation; either version 3, or (at your option) any later
\\ version.
\\
\\ This file is distributed in the hope that it will be useful, but
\\ WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
\\ or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
\\ for more details.
\\
\\ You should have received a copy of the GNU General Public License along
\\ with this file  See the file COPYING.  If not, see
\\ <http://www.gnu.org/licenses/>.


default(strictargs,1);
default(recover,0);
read("../vecper.gp");


\\ Usage: gp <Z-order-curve.gp
\\
\\ Z-order maps integers to locations in the first quadrant symbols.
\\
\\       3 |   10   11   14   15
\\         |      
\\       2 |    8    9   12   13
\\         |      
\\     Y=1 |    2 .. 3    6    7
\\         |      \
\\     Y=0 |    0 .. 1    4    5
\\         +-----------------------
\\            X=0    1    2    3
\\
\\ The X coordinate is the even bits of N and the Y coordinate are the odd
\\ bits, reckoning the least significant bit as position 0 so even.
\\
\\ This can be extended down to fractional bits too so that a rational
\\ 0 <= f <= 1 is a location somewhere in the unit square 0 <= X,Y <= 1.
\\ One way to extract the relevant bits is by taking base 4 digits with
\\ vecper_frac_digits() and modifying them to just the first or second bit
\\ each.

Zorder_f_to_x(f) =
{
  my(v=vecper_frac_digits(f,4));
  v=vecper_apply(d->bittest(d,0), v);     \\ low bit of each base-4
  vecper_frac_fromdigits(v,2);
}
Zorder_f_to_y(f) =
{
  my(v=vecper_frac_digits(f,4));
  v=vecper_apply(d->bittest(d,1), v);     \\ second bit of each base-4
  vecper_frac_fromdigits(v,2);
}

Zorder_f_to_x(0) == 0     || error();
Zorder_f_to_y(0) == 0     || error();
Zorder_f_to_x(1) == 1     || error();
Zorder_f_to_y(1) == 1     || error();

Zorder_f_to_x(5/7) == 3/7     || error();
Zorder_f_to_y(5/7) == 6/7     || error();


\\ f is a rational 0 <= f <= 1.
\\ Return its Z-order location as a complex rational x + I*y.
\\
\\ vecper_frac_fromdigits() is happy to take complex numbers as the
\\ "digits", so x and y can be combined in the vecper.
\\
f_to_z(f) =
{
  my(v=vecper_frac_digits(f,4));
  v=vecper_apply(d->bittest(d,0) + I*bittest(d,1), v);
  vecper_frac_fromdigits(v,2);
}

f_to_z(2/7) == 4/7 + 1/7*I    || error();


\\-----------------------------------------------------------------------------

\\ Coordinate plots.  Try the same with ploth() under X.
\\
\\ plot() and ploth() step f through various floating point values (t_REAL).
\\ vecper_frac_digits() accepts floats and treats them according to their
\\ underlying binary mantissa and exponent.
\\
\\ An accurate plot might want to find bounds on the x (or y) resulting from
\\ all f in a given interval so as to plot a little block of range.  Doing
\\ that in power-of-2 blocks is simplest since the nature of the bit extract
\\ means a block 1/4^k ranges is high bits and +0 to +1/2^k for the rest.
\\
plot(f=0,1, Zorder_f_to_x(f))
plot(f=0,1, Zorder_f_to_y(f))

\\-----------------------------------------------------------------------------
