\\ Copyright 2019, 2020, 2021, 2022, 2023 Kevin Ryde
\\
\\ This file is free software; you can redistribute it and/or modify it
\\ under the terms of the GNU General Public License as published by the Free
\\ Software Foundation; either version 3, or (at your option) any later
\\ version.
\\
\\ This file is distributed in the hope that it will be useful, but
\\ WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
\\ or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
\\ for more details.
\\
\\ You should have received a copy of the GNU General Public License along
\\ with this file.  See the file COPYING.  If not, see
\\ <http://www.gnu.org/licenses/>.


\\ Work in progress ... mostly runs, but going to change wildly.

\\     v = vecper_make(fixed,periodic)
\\     v = vecper_set(v,i,elem,period=0)
\\     len = vecper_length(v)
\\     len = vecper_length_oo(v)
\\     vec = vecper_vecextract_vec(v,vec)
\\     v = vecper_concat_vecs(v)
\\     v = vecper_unroll(v,n=1)
\\     v = vecper_select_state(func,v,state=0)
\\     v = vecper_ensure_fixed(v,len)
\\     v = vecper_binop(func,v1,v2)
\\     v = vecper_add(v1,v2)
\\
\\     v = vecper_contfrac(w)
\\     v = vecper_contfrac_of_xys(x,y,s)
\\     w = vecper_contfrac_to_quad(v)
\\     [x,y,s] = vecper_contfrac_to_xys(v)
\\
\\ End Summary
\\
\\ v = vecper_vecextract_vecper(v,vstep)     periodic extracts


\\-----------------------------------------------------------------------------

vecper_make(fixed,periodic) = [fixed,periodic];
{
addhelp(vecper_make,
"v = vecper_make(fixed,periodic)
Return a vecper comprising the given fixed and periodic vectors.");
}

vecper_ensure_fixed(v,len) =
{
  \\ f = #v[1] length of fixed
  \\ p = #v[2] length of periodic
  \\ grow by g = len-f
  \\ w = floor(g/p) whole copies of periodic
  \\   vector size +1 for final possible partial periodic
  \\ g%p remainder partial copy of periodic
  
  \\ print("vecper_ensure_fixed() "v" len="len);
  if(#v[1]<len,    \\  && #v[2]
     my(qr=divrem(len-#v[1], #v[2]),
        w=qr[1],                                \\ num whole copies of periodic
        vecs=vector(w+1,i, if(i<=w,v[2],[])));  \\ whole copies of periodic
     \\ print("  extend ",len-#v[1]," whole "w" remainder "qr[2]);

     if(qr[2],
        \\ partial copy of periodic, and rotate
        v[2] = concat(v[2][qr[2]+1..#v[2]],          \\ after part
                      vecs[#vecs]=v[2][1..qr[2]]));  \\ initial part

     v[1] = concat(v[1],concat(vecs)));         \\ append to fixed
  v;
}
{
addhelp(vecper_ensure_fixed,
"v = vecper_ensure_fixed(v,len)
Return vecper v with its fixed part extended if necessary to ensure
at least \"len\" many elements in that part.

This can be used for ease of inspecting or modifying some of the start of
the vector.  With the fixed part long enough, v[1][pos]=new_value can store
directly.

If the periodic part of v is empty then there is nothing to extend the fixed
part.  In this case, behaviour on asking for a len longer than the existing
fixed part is unspecified (but is currently an error).");
}

\\ NOT WORKING
vecper_set(v,i,elem,period=0) =
{
  \\ print("vecper_set() "v"  i="i" period="period);
  if(period,
     \\ set in the fixed part, if any
     while(i<=#v[1], v[1][i]=elem; i+=period);

     i -= #v[1];    \\ position 0 up in periodic part
     i %= #v[2];    \\ reduced

     if(i-#v[1] > #v[2],
        \\ print("  replicate fixed part "floor((i-#v[1])/#v[2]));
        v[1] = concat(vector(1+floor((i-#v[1])/#v[2]), i,
                             if(i==1,v[1],v[2]))));

     my(l=lcm(#v[2],period));
     \\ print("  lcm of "#v[2]" "period"  l="l);
     if(l>#v[2], v[2]=concat(vector(l/#v[2],n, v[2])));
     forstep(pos=i,#v[1],period,
        \\ print("  fixed set "pos);
        v[1][pos]=elem);
     forstep(pos=(i-#v[1]-1)%#v[2] +1, #v[2], period,
        \\ print("  periodic set "pos);
        v[2][pos]=elem);
     ,
     \\ not periodic, set only in the fixed part
     v=vecper_ensure_fixed(v,i);
     v[1][i]=elem);
  v;
}
{
addhelp(vecper_set,
"v = vecper_set(v,i,elem,period=0)
v is a vecper.  Return a copy of it with element number i set to elem.
The first element is i=1.

Optional parameter \"period\" sets elem also at positions i+period,
i+2*period, etc, repeating.

If the periodic part of v is empty then i must be within just the fixed
part, and must have period=0 no repeats.");
}



\\-----------------------------------------------------------------------------

vecper_length(v) = if(#v[2],-1,#v[1]);


vecper_length_oo(v) = if(#v[2],oo,#v[1]);
{
addhelp(vecper_length_oo,
"len = vecper_length_oo(v)
Return the length of vecper v.
If v has a non-empty periodic part then its length is infinity oo.
If v has empty periodic part then its length is length of the fixed part.");
}

vecper_vecextract_vec(v,vec) = vector(#vec,i, vecper_get(v,vec[i]));
{
addhelp(vecper_vecextract_vec,
"vec = vecper_vecextract_vec(v,vec)
v is a vecper and vec is a vector of element indices (1 upwards).
Return a vector of the elements of v at the indices in vec.
If v has empty periodic part then the indices in vec must be within the
length of the fixed part.

See vecper_get_range() to extract a range of indices [lo..hi].");
}

vecper_concat_vecs(v) =
{
  if(#v[1], v[1]=concat(v[1]));
  if(#v[2], v[2]=concat(v[2]));
  v;
}
{
addhelp(vecper_concat_vecs,
"v = vecper_concat_vecs(v)
v is a vecper whose elements are vectors.
Return a new vecper which is all those vectors concatenated, like
concat(vec) does on an ordinary vector of vectors.

Either or both fixed and periodic parts can be empty vectors, in which case
they are unchanged (unlike ordinary concat() which doesn't accept an empty
vector).");
}



\\ propagate carry ...
\\ vecper_add(v1,v2) =
\\ {
\\   v1=vecper_unroll(v1,max(0,#v2-#v1));
\\   v2=vecper_unroll(v2,max(0,#v1-#v2));
\\   if(v1[#v1]==[],v1[#v1]=[0]);
\\   if(v2[#v2]==[],v2[#v2]=[0]);
\\   my(len=lcm(#v1[#v1],#v2[#v2]));
\\   v1[#v1]=concat(vector(len/#v1[#v1],i,v1[#v1]));
\\   v2[#v2]=concat(vector(len/#v2[#v2],i,v2[#v2]));
\\   vecper_normalize(v1+v2);
\\ }

vecper_unroll(v,n=1) =
{
  \\ ENHANCE-ME: n divrem periodic length
  for(i=1,n,
     my(p=v[#v]);
     v=concat([ v[^-1],
               [if(#p,p[1],0)],
               [vecper_INTERNAL_vector_rotate_down(p)]]));
  v;
}

\\ v is a vector.  Return a "rotated down" [ v[2], v[3], ..., v[#v], v[1] ].
\\ If v is less than 2 elements then return unchanged.
vecper_INTERNAL_vector_rotate_down(v) = \
  if(#v<2,v, concat([v[2..-1],v[1]]));

\\ Like digits(n,base), but pad the result to be exactly k digits.
\\ If n is more than k digits then throw an error.
vecper_INTERNAL_digits_padded(n,k,base=10) =
{
  my(v = digits(n,base));
  if(#v>k, error("vecper_INTERNAL_digits_padded() n=",n,
                 " is bigger than k=",k," digits"));
  concat(vector(max(0,k-#v)),v);
}

vecper_select_state(func,v,state=0) =
{
  vecper_concat_vecs(vecper_apply_state((elem,state)->
                                        my(ret=func(elem,state));
                                        [if(ret[1],[elem],[]),
                                         state],
                                        v,
                                        state));
}
{
addhelp(vecper_select_state,
"v = vecper_select_state(func,v,state=0)
Select elements from v according to func and a \"state\" maintained, and
return the result as a new vecper.

func is a predicate function called on successive elements of v from i=1
upwards.  It should return a 2-element vector

  [bool,new_state] = func(elem,state)

state is initially the given state parameter, then the successive new_state.
If bool is non-0 then elem is kept in the return.

When a repeat state is found after the periodic part, that determines the
periodic part of the return.  \"state\" values can be anything, but must
eventually repeat on the periodic part.  States are compared by \"==\" for
identifying a repeat.");
}

vecper_binop(func,v1,v2) =
{
  my(fixed=vector(max(#v1[1],#v2[1])));
  \\ for(i=1,#fixed
  \\                 func(vecper_get(v1,i),vecper_get(v2,i))));
  if(#v1[2]==0,
     if(#v2[2]==0,
        \\ both fixed length
        v1[1]=v2[1];
        ,
        v1[2]=vecper_get_range(v2,#fixed+1,#fixed+#v2[2])),
     v2[2]=vecper_get_range(v1,#fixed+1,#fixed+#v1[2]);
     ,
     v2[2]=vector(lcm(#v1[2],#v2[2]),i,
                  func(vecper_get(v1,i+#fixed),vecper_get(v2,i+#fixed))));
  v1[1]=fixed;
  v1;
}
{
addhelp(vecper_binop,
"v = vecper_binop(func,v1,v2)
Return a vecper which is the term-wise binary operator applied to vecpers v1
and v2 as v[n] = func(v1[n], v2[n]).");
}

vecper_add(v1,v2) = vecper_binop((x,y)->x+y, v1,v2);


\\-----------------------------------------------------------------------------
\\ Continued Fractions

\\ x + y*sqrt(s) 
vecper_INTERNAL_floor_of_quad(q) =
{
  my(den=denominator(q));
  q*=den;
  (real(q) + sqrtint(q.disc*sqr(imag(q)))\2) \ den;
}

vecper_contfrac(w) =
{
  if(type(w)=="t_QUAD",
     if(bitand(w.disc,1),
        vecper_contfrac_of_xys((real(w)+imag(w))/2, imag(w)/2, w.disc),
        vecper_contfrac_of_xys(real(w), imag(w), w.disc/4)),
     [contfrac(w),[]]);

  \\ if(type(w)=="t_QUAD",
  \\    vecper_make_state((w)->if(w, my(t=vecper_INTERNAL_floor_of_quad(w)); [t,if(w-=t,1/w)], 'end), w),
  \\    [contfrac(w),[]]);
}
{
addhelp(vecper_contfrac,
"v = vecper_contfrac(w)
Return the continued fraction expansion of w in the form of a vecper.

If w is rational or floating point then this is a finite-length vecper
equivalent to what contfrac() gives.
If w is a real quad (\"t_QUAD\") then the continued fraction is eventually
periodic.
See vecper_contfrac_of_xys() for notes on the nature of the vecper.");
}

vecper_contfrac_of_xys(x,y,s) =
{
  \\ print("vecper_contfrac_of_xys() "x" + "y" * sqrt("s")");

  \\ w = x + y*sqrt(abs(s))   where y = +/-1
  \\
  \\ If s is a perfect square then w is rational and the same finite
  \\ sequence of terms results no matter how x and s together give w.
  \\
  \\ If s is not a perfect square then this representation and value w
  \\ are one-to-one.  Every value w has a distinct continued fraction, so
  \\ "xs" is a unique key and gives the minimal vecper.
  \\
  \\ denominator() gives 1 when an arg is 0, unlike lcm(x,s) which will
  \\ give 0.
  \\ 
  vecper_make_state
  ((xys)->
   \\ print("  xys="xys"  approx ",x+y*sqrt(s));
   if(xys=='end,
      'end,
      my(x=xys[1],
         y=xys[2],
         s=xys[3],
         den=denominator(xys),
         t=floor((x*den + y*sqrtint(s*sqr(den))) / den));
      x -= t;
      my(n=sqr(x)-s);
      \\ print("  t="t" to x="x"  new norm="n);
      if(n,
         [t, [x/n, -y*sign(n), s/sqr(n)]];
         ,
         [t, 'end])),
   [x,sign(y),s*sqr(y)]);
}
{
addhelp(vecper_contfrac_of_xys,
"v = vecper_contfrac_of_xys(x,y,s)
Return a vecper which is the continued fraction expansion of w = x+y*sqrt(s).
x,y,s should be rationals, and must have s>=0.

If w is rational (y=0 or s perfect square), then the return is a
fixed-length vecper equivalent to what contfrac() gives.
Otherwise, w is a quadratic irrational and the continued fraction is
eventually periodic.

w=0 is returned as a single 0 term the same as contfrac() gives, so always
length >= 1.  If w<0 then the integer term is negative.  All further terms
are >0.  The return is minimal in all cases.");
}

\\ a + b*sqrt(s)
\\ 1/(a + b*sqrt(s)) = (a - b*sqrt(s)) / (a^2 - b^2*s)
vecper_contfrac_to_xys(v) =
{
  \\ print("vecper_contfrac_to_xys() "v);
  \\ if(v[2],
  \\    my(m=contfracpnqn(v[2]));
  \\    \\ print("m="m);
  \\    my(a=2*m[2,1],
  \\       b=m[2,2] - m[1,1],
  \\       c = -m[1,2],
  \\       s = sqr(b)-2*a*c);
  \\    \\ print("a="a/2" b="b" c="c" s="s);
  \\    my(x=-b/a, y=1/a);
  \\    \\ print("  to x="x" y="y" s="s);
  \\    forstep(i=#v[1],1,-1,     \\ x + y*sqrt(s)
  \\       my(n=x^2 - y^2*s);
  \\       x = v[1][i] + x/n;
  \\       y/=-n);
  \\    [x,y,s];
  \\    \\ m=contfracpnqn(v[1]);
  \\    \\ print("  fixed m="m);
  \\    \\ [ m[1,1]*x + m[1,2]*y,
  \\    \\   m[2,1]*x + m[2,2]*y,
  \\    \\   s ];
  \\    ,
  \\    my(m=contfracpnqn(v[1]));
  \\    [m[1,1]/m[2,1], 0, 0]);


  \\     r*a + b
  \\ r = -------
  \\     r*c + d
  \\
  \\ r^2*c + (d-a)*r - b = 0
  \\ s = (d-a)^2 + 4*b*c
  \\      a-d    1
  \\ r =  --- + --- * sqrt(s)
  \\      2*c   2*c
  \\     \---/ \---/
  \\       x     y

  if(v[2],
     my(m=contfracpnqn(v[2]));
     \\ print("  periodic m="m);
     my(A=2*m[2,1],
        B = m[1,1] - m[2,2],
        C = -m[1,2],
        s = sqr(B)-2*A*C);
     \\ print("    A="A/2" B=",-B," C="C" s="s);
     my(x=B/A, y=1/A);
     \\ print("    to x="x" y="y" s="s);
  
     \\ alpha = x+y*sqrt(s)
     \\       alpha*a + b
     \\ ret = -----------
     \\       alpha*c + d
     \\
     \\ den  = (x+y*sqrt(s))*c + d
     \\      = x*c+d + y*c*sqrt(s)
     \\ num  = x*a+b + y*a*sqrt(s)
     \\ denC = x*c+d - y*c*sqrt(s)
     \\
     \\ (x*c+d)*(x*a+b) - y*c*y*a*s  == x^2*c*a + x*(a*d + c*b) + b*d - y^2*c*a*s  
     \\                               = (x^2+s*y^2)*a*c  + x*(a*d + b*c) + b*d
     \\
     \\ (x*c+d)*y*a - (x*a+b)*y*c == x*y*a*c + y*a*d  - x*y*a*c - y*b*c
     \\                            = y*(a*d - b*c)
     \\
     \\ norm(den) = (x*c+d)^2 - y^2*s*c^2
     \\           

     m=contfracpnqn(v[1]);
     my(a=m[1,1],
        b=m[1,2],
        c=m[2,1],
        d=m[2,2],
        den_x=x*c+d,
        N = sqr(den_x) - sqr(y*c)*s);
     
     [(den_x*(x*a+b) - sqr(y)*a*c*s) / N,
      y*(a*d - b*c) / N,
      s];
     ,
     my(m=contfracpnqn(v[1]));
     [m[1,1]/m[2,1], 0, 0]);
     
  \\ A = x*m[1,1] + m[1,2];
  \\ B = y*m[1,1];
  \\ x = x*m[2,1] + m[2,2];
  \\ y = y*m[2,1];
  \\ C = x^2 - y^2*s;
  \\ 
  \\ [(A*x + B*y*s)/C,
  \\  (A*y + B*x)/C,
  \\  s];
}
{
addhelp(vecper_contfrac_to_xys,
"[x,y,s] = vecper_contfrac_to_xys(v)
v is a vecper which is terms of a continued fraction.
Return a vector of 3 elements [x,y,s] which is its value x + y*sqrt(s).
Each of x,y,s are rationals.

If v is finite or the periodic part all 0s, then y=0 and s=0 and the value
is rational x.

If v has a non-zero periodic part, then the value is a quadratic irrational
which is expressed in some combination of rationals x,y,s.  But exactly
which y and s is unspecified.  It's possible that s is not square-free (some
of it might come out into y).");
}

vecper_contfrac_to_quad(v) =
{
  my(xys=vecper_contfrac_to_xys(v),
     y=xys[2],
     s=xys[3]);
  xys[1]
  + if(y&&s,
       my(c=denominator(s));
       y/=c;
       s*=sqr(c); \\ s integer
       c=core(s,1);  \\ [squarefree, sqrtofsquare]
       s=c[1];
       y*=c[2];
       if(s==1,y,y*quadgen(4*s)));
}
{
addhelp(vecper_contfrac_to_quad,
"w = vecper_contfrac_to_quad(v)
v is a vecper which is terms of a continued fraction.

Return its value in the form of a rational or a real quad (\"t_QUAD\").
Any fixed-length v is rational, and some periodic v are rational.

Most periodic v are quadratic irrationals and they are returned as a real
quad using discrminant = 0 mod 4 (not the 1 mod 4 option available to some
roots).

The current implementation finds the necessary square-free part discriminant
using core().  This is a factorization and might be unusably slow on big
values.");
}

\\-----------------------------------------------------------------------------

\\ Local variables:
\\ compile-command: "gp -f --quiet <test-wip.gp"
\\ End:

\\  LocalWords:  vecper elem len vec vecextract concat vecs func bool xys
\\  LocalWords:  contfrac normalized num contfracpnqn factorize unusably
